/***************************************************************************
 * DIRFILE.C -                                                             *
 * -----------------------                                                 *
 *                                                                         *
 * VCDID - CD-I filesystem driver                                          *
 * (c) 1997 - 1999 Vahur Sinijrv                                          *
 *                                                                         *
 * You are free to use any part of this source code, provided              *
 * that credit is given to author in such cases.                           *
 ***************************************************************************/
 
#include <basedef.h>
#include <vmm.h>
#include <error.h>
#include "vcdid.h"

#pragma VxD_PAGEABLE_CODE_SEG
#pragma VxD_PAGEABLE_DATA_SEG

DWORD DirFile_Init( PDIRFILE pDirFile, LPCDIFS_VOLUME l, ParsedPath *pPath )
{
    DWORD       dwDirSect;
    DWORD       dwResult;
    PCDI_DIRENT pDirEnt;

    pDirFile->pBuffer = NULL;

    dwDirSect = FindDirRec( l, pPath );
    if ( !dwDirSect )
        return ERROR_PATH_NOT_FOUND;

    dwResult = IO_GetCachePtr( l, &pDirEnt, dwDirSect, 0 );
    if ( dwResult != NO_ERROR )
        return dwResult;

    return DirFile_InitFromCache( pDirFile, l, pDirEnt );
}

DWORD DirFile_InitFromCache( PDIRFILE       pDirFile,
                             LPCDIFS_VOLUME l,
                             PCDI_DIRENT    pDirEnt )
{
    DWORD       dwSize;
    int         nOffs;
    DWORD       dwBufSize;

    dwSize    = SWAPD( pDirEnt->size );
    dwBufSize = dwSize < COOKED_SECT_SIZE ? dwSize : COOKED_SECT_SIZE;

    pDirFile->pBuffer = (LPCDI_DIRENT)_HeapAllocate( dwBufSize, HEAPSWAP );
    if ( !pDirFile->pBuffer )
        return ERROR_GEN_FAILURE;

    pDirFile->nTotal  = dwSize;
    pDirFile->dwLBN   = SWAPD( pDirEnt->lbn );
    pDirFile->pVolume = l;

    /* Skip over . directory */
    pDirFile->nOffs = pDirEnt->length;

    dwSize = IO_ReadCachedBytes( l, pDirFile->pBuffer, pDirFile->dwLBN, dwBufSize );
    if ( dwSize != NO_ERROR )
    {
        _HeapFree( pDirFile->pBuffer, 0 );
        pDirFile->pBuffer = NULL;
        return dwSize;
    }

    return NO_ERROR;
}

PCDI_DIRENT DirFile_GetNextEntry( PDIRFILE pDirFile )
{
    int         iSize;
    DWORD       dwResult;
    PCDI_DIRENT pCurrent;

    if ( pDirFile->nOffs >= pDirFile->nTotal )
        return NULL;

    pCurrent = (PCDI_DIRENT)((char*)pDirFile->pBuffer + pDirFile->nOffs);

    pDirFile->nOffs += pCurrent->length;
    pCurrent         = PDIRENT_NEXT( pCurrent );

    /* End of directory entries */
    if ( pDirFile->nOffs >= pDirFile->nTotal )
        return NULL;

    /* Violation of CD-I specs */
    if ( pDirFile->nOffs > COOKED_SECT_SIZE )
    {
        pDirFile->nTotal = 0;
        return NULL;
    }

    /* Next sector */
    if ( pDirFile->nOffs == COOKED_SECT_SIZE ||
         pCurrent->length == 0 )
    {
        pDirFile->nTotal -= COOKED_SECT_SIZE;

        iSize = ( pDirFile->nTotal > COOKED_SECT_SIZE ) ?
            COOKED_SECT_SIZE :
            pDirFile->nTotal;

        dwResult = IO_ReadCachedBytes(
            pDirFile->pVolume,
            pDirFile->pBuffer,
            ++(pDirFile->dwLBN),
            iSize );

        if ( dwResult != NO_ERROR )
        {
            pDirFile->nTotal = 0;
            return NULL;
        }

        pDirFile->nOffs = 0;
        pCurrent = pDirFile->pBuffer;
    }

    /* Partial entry */
    if ( pDirFile->nOffs + pCurrent->length > pDirFile->nTotal )
    {
        pDirFile->nTotal = 0;
        return NULL;
    }

    return pCurrent;
}

void DirFile_Done( PDIRFILE pDirFile )
{
    if ( pDirFile->pBuffer )
    {
        _HeapFree( pDirFile->pBuffer, 0 );
        pDirFile->pBuffer = NULL;
    }
}

