#ifndef VCDID_H
#define VCDID_H

#include <dcb.h>
#include <vrp.h>
#include <ior.h>
#include <ios.h>
#include "sys/ios_vxd.h"
#include <ifs.h>
#include "sys/ifsmgr_vxd.h"
#include <srb.h>
#include <vcache.h>
#include "cdi_fs.h"

#ifdef DEBUG
#define NEEDS_MEMCPY
#endif

#define COOKED_SECT_SIZE    2048
#define RAW_SECT_SIZE       2352
#define SECT_SIZE           2324
#define SYNC_SIZE           24

typedef struct tagFileInfo {
    struct tagFileInfo *pNext, *pPrev;
    DWORD   	       start;
    DWORD              length;
    DWORD              attr;
    DWORD              pos;
    dos_time           creation_time;
    unsigned short     creation_10ms;
    BYTE               *buf;
    DWORD              bufsect;
} FILEINFO, *LPFILEINFO, *PFILEINFO;

#pragma pack( push, 1 )
typedef struct _CDIFS_VOLUME {
    struct _CDIFS_VOLUME *next;
    ULONG            media_handle;
    DCB              *dcb;
    VRP              *vrp;
    BYTE             drive;
    DWORD            ptbl_size;                 /* size of path table       */
    LPCDI_PATH_TABLE ptbl;                      /* ptr to path table        */
    WORD             size;                      /* size of volume in blocks */
    DWORD            serial;                    /* serial number            */
    LPFILEINFO       pFileList;                 /* list of opened files     */
    DWORD            nFiles;                    /* number of open files     */
    DWORD            label_size;
    unsigned short   label[ MAX_CDI_VOL_ID ];
    DWORD            volume_id_length;
    unsigned char    volume_id[ MAX_CDI_VOL_ID ];
    DWORD            album_id_length;
    unsigned char    album_id[ MAX_CDI_ALBUM_ID ];
    DWORD            publisher_id_length;
    unsigned char    publisher_id[ MAX_CDI_PUBLISHER_ID ];
    DWORD            preparer_id_length;
    unsigned char    preparer_id[ MAX_CDI_PREPARER_ID ];
    DWORD            application_id_length;
    unsigned char    application_id[ MAX_CDI_APPLICATION_ID ];
    unsigned char    creation_date[ MAX_CDI_DATETIME ];
    unsigned char    modification_date[ MAX_CDI_DATETIME ];
    unsigned char    expiration_date[ MAX_CDI_DATETIME ];
    unsigned char    effective_date[ MAX_CDI_DATETIME ];
    DWORD            volume_set_size;
    DWORD            volume_sequence_number;
} CDIFS_VOLUME, *LPCDIFS_VOLUME;
#pragma pack( pop )

typedef struct
{
    int            nOffs;
    int            nTotal;
    PCDI_DIRENT    pBuffer;
    LPCDIFS_VOLUME pVolume;
    DWORD          dwLBN;
} DIRFILE, *LPDIRFILE, *PDIRFILE;

/* Globals */
struct volfunc cdifs_volfunc;
struct hndlmisc cdifs_hndlmisc;
struct hndlmisc cdifs_hndlopen;

extern LPCDIFS_VOLUME vol_list;
extern BYTE           g_CacheID;

/*                   */
/* Assembler helpers */
/*                   */
DCB*      _cdecl GetPhysDCB( BYTE drive );
IOR*      _cdecl AllocIOP( LPCDIFS_VOLUME );
void      _cdecl FreeIOP( IOR * );

#define IO_READBLOCK_RAW    1
DWORD     IO_Read( LPCDIFS_VOLUME pVol, void* pBuf, DWORD dwBlock, DWORD dwCount, DWORD dwFlags );
DWORD     IO_ReadCached( LPCDIFS_VOLUME pVol, void* pBuf, DWORD dwBlock, DWORD dwFlags );
DWORD     IO_GetCachePtr( LPCDIFS_VOLUME pVol, void** pPtr, DWORD dwBlock, DWORD dwFlags );
DWORD     IO_ReadCachedBytes( LPCDIFS_VOLUME pVol, void* pBuf, DWORD dwBlock, DWORD dwBytes );

void      _cdecl IO_ReleaseCachePages( LPCDIFS_VOLUME );
DWORD     _cdecl MulDiv( DWORD x, DWORD y, DWORD z );
int       _cdecl strncmp( char *s1, char *s2, int count );
unsigned  _cdecl trim_right( char *s, int len );
void      _cdecl zeromem( void *, int );

#ifdef NEEDS_MEMCPY
/* Surprise, surprise - VC50 has built-in memcpy - how odd ... */
int  _cdecl _memcpy( void *dest, void *src, int count );
#endif

/* Flags for IO_ReadBlock */
#define IO_READF_COOKED 0
#define IO_READF_RAW    1

DWORD _cdecl CDIFS_Close( pioreq p );
DWORD _cdecl CDIFS_Dir( pioreq p );
DWORD _cdecl CDIFS_DirectDiskIO( pioreq p );
DWORD _cdecl CDIFS_EnumerateHandle( pioreq p );
DWORD _cdecl CDIFS_Fail( pioreq p );
DWORD _cdecl CDIFS_FailWRP( pioreq p );
DWORD _cdecl CDIFS_FileAttributes( pioreq p );
DWORD _cdecl CDIFS_FileLocks( pioreq p );
DWORD _cdecl CDIFS_FileTimes( pioreq p );
DWORD _cdecl CDIFS_FindFirstFile( pioreq p );
DWORD _cdecl CDIFS_FindNextFile( pioreq p );
DWORD _cdecl CDIFS_FindClose( pioreq p );
DWORD _cdecl CDIFS_HandleInfo( pioreq p );
DWORD _cdecl CDIFS_IOCtl16( pioreq p );
DWORD _cdecl CDIFS_MountVolume( pioreq p );
DWORD _cdecl CDIFS_OpenFile( pioreq p );
DWORD _cdecl CDIFS_QueryResourceInfo( pioreq p );
DWORD _cdecl CDIFS_ReadFile( pioreq p );
DWORD _cdecl CDIFS_Seek( pioreq p );
DWORD _cdecl CDIFS_Unmount( pioreq p );
DWORD _cdecl CDIFS_GetDiskInfo( pioreq p );

/*
 *  Utility functions
 */
DWORD          FindDirRec( LPCDIFS_VOLUME, ParsedPath * );
unsigned short CDIToDosTime( BYTE* , dos_time* );
DWORD          DirFile_Init( PDIRFILE, LPCDIFS_VOLUME, ParsedPath* );
DWORD          DirFile_InitFromCache( PDIRFILE, LPCDIFS_VOLUME, PCDI_DIRENT );
void           DirFile_Done( PDIRFILE );
PCDI_DIRENT    DirFile_GetNextEntry( PDIRFILE );

/*
 *  Utility inlines
 */
DWORD __inline
swapd( DWORD i )
{
    __asm mov   eax, i
    __asm bswap eax
}

WORD __inline
swapw( WORD i )
{
    __asm mov   ax, i
    __asm xchg  ah, al
}

/*
 *  Few C callable routines
 */

/*
 * VCache
 */
typedef struct
{
    DWORD   dw0;
    DWORD   dw4;
    DWORD   FSKey1;
    DWORD   FSKey2;
    DWORD   BufPtr;
    DWORD   FSDData[ 7 ];
} VCACHE_HANDLE, *PVCACHE_HANDLE;

PVCACHE_HANDLE VXDINLINE
VCache_FindBlock( BYTE option, BYTE fsid, DWORD key1, DWORD key2 )
{
    Touch_Register( eax );
    Touch_Register( esi );
    __asm mov al, option
    __asm mov ah, fsid
    __asm mov ebx, key1
    __asm mov edi, key2
    VxDCall( VCache_FindBlock );
    __asm mov eax, esi
}

/*
 * IOS
 */
DWORD VXDINLINE
_IOS_SendCommand( IOR *ior, DCB *dcb )
{
    DCB_BLOCKDEV *bd;

    bd = &dcb->DCB_bdd;
    Touch_Register( eax );
    Touch_Register( ebx );
    Touch_Register( ecx );
    Touch_Register( edx );
    __asm mov esi, [ior]
    __asm mov edi, [bd]
    VxDCall( IOS_SendCommand );
}

_QWORD VXDINLINE
IFSMgr_BCSToUNI( string_t s1, unsigned char* s2, unsigned int count, int charset )
{
    Touch_Register( eax );
    Touch_Register( ecx );
    Touch_Register( edx );
    __asm push charset
    __asm push count
    __asm push s2
    __asm push s1
    VxDCall( BCSToUni );
    __asm add  esp, 16
}

unsigned int VXDINLINE
IFSMgr_MetaMatch( string_t s1, string_t s2, int matchSem )
{
    Touch_Register( eax );
    Touch_Register( ecx );
    Touch_Register( edx );
    __asm push matchSem
    __asm push s2
    __asm push s1
    VxDCall( IFSMgr_MetaMatch );
    __asm add  esp, 12
}

_FILETIME VXDINLINE
IFSMgr_DosToWin32Time( dos_time dtime )
{
    __asm push dtime
    VxDCall( IFSMgr_DosToWin32Time );
    __asm add esp, 4
}
#endif
