/***************************************************************************
 * IOCTL.C - IO Control functions ( accessible thru DeviceIoControl )      *
 * ------------------------------------------------------------------      *
 *                                                                         *
 * VCDID - CD-I filesystem driver                                          *
 * (c) 1997 - 1999 Vahur Sinijrv                                          *
 *                                                                         *
 * You are free to use any part of this source code, provided              *
 * that credit is given to author in such cases.                           *
 ***************************************************************************/
 
#include <basedef.h>
#include <vmm.h>
#include <vwin32.h>
#include <winerror.h>
#include "ioctl.h"
#include "vcdid.h"
#include "cfg.h"

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG

#define COPY_ID( dst, src, field )  \
    memcpy( dst->field, src->field, src->field##_length ); \
    dst->field[ src->field##_length ] = '\0'

#define COPY_DATE( dst, src, field ) \
    memcpy( dst->field, src->field, sizeof( src->field ) )

typedef DIOCPARAMETERS* LPDIOC;
typedef DWORD*          LPDWORD;

/*
 * Helpers
 */
DWORD GetMountedVolumesCount()
{
    int i;
    LPCDIFS_VOLUME pVol;

    pVol = vol_list;
    for( i = 0; pVol; pVol = pVol->next ) i++;
    return i;
}

/*
 * IOCTL functions
 */

//----------------------------------------------------------------------
// Name: IOC_GET_MOUNT_COUNT
//
// Desc: Return number of mounted volumes. DeviceIoControl call's output
//       buffer must point to dword which gets the return value.
//----------------------------------------------------------------------
DWORD _stdcall IOC_GET_MOUNT_COUNT( LPDIOC p )
{
    DWORD *pdw;

    pdw = (DWORD*)p->lpvOutBuffer;
    *pdw = GetMountedVolumesCount();
    return 0;
}

//----------------------------------------------------------------------
// Name: IOC_GET_MOUNTED_VOLUMES
//
// Desc: Get descriptions of mounted volumes. Input buffer has to point
//       to dword, which specifes how many IOCS_CDIFS_VOLUME structures
//       are present in output buffer. Output buffer is a pointer to
//       IOCS_CDIFS_VOLUMES structure which will be filled with volume
//       descriptions. dwVolumes member of this structure cointains the
//       number of IOCS_CDIFS_VOLUME's filled, it may differ from
//       previous IOC_GET_MOUNT_COUNT call.
//----------------------------------------------------------------------
DWORD _stdcall IOC_GET_MOUNTED_VOLUMES( LPDIOC p )
{
    PIOCS_CDIFS_VOLUMES pVolumes;
    PIOCS_CDIFS_VOLUME  pVolume;
    LPCDIFS_VOLUME      pVol;
    DWORD               nVolumes;
    DWORD               i;

    pVolumes  = (PIOCS_CDIFS_VOLUMES)p->lpvOutBuffer;
    pVolume   = pVolumes->volumes;
    nVolumes  = *(DWORD*)p->lpvInBuffer;

    pVol = vol_list;
    for( i = 0; i < nVolumes; i++, pVolume++ )
    {
        if ( pVol )
        {
            COPY_ID( pVolume, pVol, volume_id );
            COPY_ID( pVolume, pVol, album_id );
            COPY_ID( pVolume, pVol, publisher_id );
            COPY_ID( pVolume, pVol, preparer_id );
            COPY_ID( pVolume, pVol, application_id );
            COPY_DATE( pVolume, pVol, creation_date );
            COPY_DATE( pVolume, pVol, modification_date );
            COPY_DATE( pVolume, pVol, expiration_date );
            COPY_DATE( pVolume, pVol, effective_date );
            pVolume->volume_set_size = pVol->volume_set_size;
            pVolume->volume_sequence_number = pVol->volume_sequence_number;
            pVolume->nFiles = pVol->nFiles;
            pVolume->drive  = pVol->drive;
        }
        else
        {
            break;
        }

        pVol = pVol->next;
    }

    pVolumes->dwVolumes = i;
    return ERROR_SUCCESS;
}

DWORD _stdcall IOC_SET_CONFIG( LPDIOC p )
{
    CDIFS_CONFIG* pConfig;

    if ( p->cbInBuffer != sizeof( CDIFS_CONFIG ) )
    {
        return ERROR_INVALID_PARAMETER;
    }

    pConfig = (CDIFS_CONFIG*)p->lpvInBuffer;
    return config_set( pConfig );
}

DWORD _stdcall IOC_GET_CONFIG( LPDIOC p )
{
    CDIFS_CONFIG* pConfig;

    if ( p->cbOutBuffer != sizeof( CDIFS_CONFIG ) )
    {
        return ERROR_INVALID_PARAMETER;
    }

    pConfig = (CDIFS_CONFIG*)p->lpvOutBuffer;
    memcpy( pConfig, &g_config, sizeof( CDIFS_CONFIG ) );
    return ERROR_SUCCESS;
}
